use iolaus::data_types::{FormulaQuestion, PlainQuestion, RawQuestion, Variable};
use iolaus::utils;
use std::collections::HashMap;
use std::fs;
use std::fs::DirEntry;
use std::path::Path;
use std::str::FromStr;
use yaml_rust::{Yaml, YamlLoader};

// Read folder and loop through its files with extension yml
pub fn get(dir_path: &Path, n_alternatives: u8) -> Vec<RawQuestion> {
    let entries: Vec<DirEntry> = match fs::read_dir(dir_path) {
        Err(_) => panic!(
            "Erro ao ler a pasta: {}",
            dir_path.to_str().expect("Erro #0301")
        ),
        Ok(entries) => entries
            .map(|e| e.expect("Erro #0302"))
            .filter(|e| e.file_type().expect("Erro #0303").is_file())
            .filter(|e| e.path().extension().expect("Erro #0304") == "yml")
            .collect(),
    };
    let mut questions = Vec::new();
    for entry in entries {
        let raw_question = parse(&entry.path(), n_alternatives);
        questions.push(raw_question);
    }
    questions
}

// Read file contents and parse questions
fn parse(path: &Path, n_alternatives: u8) -> RawQuestion {
    let file_path = path.to_str().expect("Erro #0305");
    let content = utils::read_file_content(path);
    let docs = match YamlLoader::load_from_str(&content) {
        Err(_) => panic!("{}: formato YML inválido", file_path),
        Ok(docs) => docs,
    };
    let doc = &docs[0];
    match doc["alternativas"].is_badvalue() {
        true => parse_formula_question(doc, file_path),
        false => parse_plain_question(doc, file_path, n_alternatives),
    }
}

fn parse_plain_question(doc: &Yaml, file_path: &str, n_alternatives: u8) -> RawQuestion {
    RawQuestion::Plain(PlainQuestion {
        stem: parse_string("enunciado", doc, file_path),
        answer: parse_string("resposta", doc, file_path),
        alternatives: parse_alternatives(doc, file_path, n_alternatives),
    })
}

fn parse_formula_question(doc: &Yaml, file_path: &str) -> RawQuestion {
    RawQuestion::Formula(FormulaQuestion {
        stem: parse_string("enunciado", doc, file_path),
        answer: parse_string("resposta", doc, file_path),
        variables: parse_variables(doc, file_path),
        parameters: parse_parameters(doc, file_path),
    })
}

fn parse_string(parameter: &str, doc: &Yaml, file_path: &str) -> String {
    match doc[parameter].as_str() {
        Some(item) => item.to_string(),
        None => panic!("{}: {} não especificado(a)", file_path, parameter),
    }
}

fn parse_variables(doc: &Yaml, file_path: &str) -> Vec<Variable> {
    match &doc["variáveis"] {
        Yaml::Hash(yaml_hash) => {
            let mut variables = Vec::new();
            for (key, node) in yaml_hash.iter() {
                let name = match key.as_str() {
                    Some(x) => x.to_string(),
                    None => panic!("{}->variáveis: nome inválido", file_path),
                };
                let min = parse_variables_range("min", &name, node, file_path);
                let max = parse_variables_range("max", &name, node, file_path);
                variables.push(Variable { name, min, max });
            }
            variables
        }
        _ => panic!("{}: erro ao ler variáveis", file_path),
    }
}

fn parse_variables_range(range: &str, var: &String, node: &Yaml, file_path: &str) -> i32 {
    match node[range].as_i64() {
        Some(x) => x as i32,
        None => panic!(
            "{}->variáveis->{}->{}: valor inválido. o valor é um inteiro?",
            file_path, var, range
        ),
    }
}

fn parse_alternatives(doc: &Yaml, file_path: &str, n_alternatives: u8) -> Vec<String> {
    let alternatives: Vec<String> = match &doc["alternativas"] {
        Yaml::Array(arr) => arr
            .iter()
            .map(|item| match item.as_str() {
                Some(x) => x.to_owned(),
                None => panic!("{}: alternativa inválida", file_path),
            })
            .collect(),
        _ => panic!("{}: alternativas não é uma lista", file_path),
    };
    if alternatives.len() < (n_alternatives - 1) as usize {
        panic!(
            "{}: número insuficiente de alternativas. espeficique pelo menos {}",
            file_path,
            n_alternatives - 1
        );
    }
    alternatives
}

fn parse_parameters(doc: &Yaml, file_path: &str) -> HashMap<String, f32> {
    match &doc["parâmetros"] {
        Yaml::Hash(yaml_hash) => {
            let mut hash_map = HashMap::new();
            for (k, v) in yaml_hash.iter() {
                let key = match k.as_str() {
                    Some(x) => x.to_string(),
                    None => panic!("{}->parâmetros: nome inválido", file_path),
                };
                let value = match v {
                    Yaml::Integer(v) => {
                        if key == "casas_decimais" {
                            v.clone() as f32
                        } else if v > &0i64 {
                            v.clone() as f32
                        } else {
                            panic!("{}->parâmetro->{}: valor inválido", file_path, key);
                        }
                    }
                    Yaml::Real(v) => {
                        if key == "delta_min" || key == "delta_max" {
                            let vv = f32::from_str(v).expect("Erro #0306");
                            if vv > 0f32 {
                                vv
                            } else {
                                panic!("{}->parâmetro->{}: valor inválido", file_path, key);
                            }
                        } else {
                            panic!("{}->parâmetro->{}: valor inválido", file_path, key);
                        }
                    }
                    _ => panic!("{}->parâmetro->{}: valor inválido", file_path, key),
                };
                hash_map.insert(key, value);
            }
            hash_map
        }
        _ => panic!("{}: erro ao ler parâmetros", file_path),
    }
}
