use iolaus::data_types::{Config, FormulaQuestion, Question};
use rand::{thread_rng, Rng};
use std::process::Command;

pub fn process(raw_question: &FormulaQuestion, config: &Config) -> Question {
    let decimals = raw_question
        .parameters
        .get("casas_decimais")
        .unwrap_or(&0f32)
        .clone() as u8;

    let variables = get_variables(raw_question, decimals);
    let answer_value = get_answer(raw_question, &variables, config);
    let (raw_alternatives, answer) =
        get_alternatives(raw_question, answer_value, decimals, config.n_alternatives);

    Question {
        stem: format_stem(raw_question, &variables),
        alternatives: format_alternatives(raw_alternatives, &variables),
        answer,
    }
}

fn format_stem(raw_question: &FormulaQuestion, variables: &Vec<(String, String)>) -> String {
    let mut stem = raw_question.stem.clone();
    for var in variables {
        let r = format!("@{{{}}}", &var.0);
        stem = stem.replace(&r, &var.1);
    }
    stem
}

fn format_alternatives(alternatives: Vec<f32>, variables: &Vec<(String, String)>) -> Vec<String> {
    let significant = variables.iter().fold(255, |min, x| {
        let val = get_significant_figure(&x.1);
        if val < min {
            val
        } else {
            min
        }
    });
    alternatives
        .iter()
        .map(|x| format!("{:.10}", x))
        .map(|s| format_significant_figure(s, significant))
        .collect()
}

fn format_significant_figure(number: String, significant: u8) -> String {
    let mut counter = 0u8;
    let mut index = 0u8;
    let mut started = false;
    let mut dot_found = false;
    let mut truncate_dot: bool;
    for c in number.chars() {
        truncate_dot = false;
        match c {
            '0' => if started {
                counter = counter + 1;
            },
            '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9' => {
                started = true;
                counter = counter + 1;
            }
            '.' => {
                dot_found = true;
                truncate_dot = true;
            }
            _ => (),
        };
        index = index + 1;
        if counter >= significant && dot_found {
            if truncate_dot {
                index = index - 1
            };
            break;
        };
    }
    let mut formatted = number.clone();
    formatted.truncate(index as usize);
    formatted
}

fn get_significant_figure(number: &String) -> u8 {
    let mut counter = 0u8;
    let mut started = false;
    for c in number.chars() {
        match c {
            '0' => if started {
                counter = counter + 1;
            },
            '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9' => {
                started = true;
                counter = counter + 1;
            }
            _ => (),
        }
    }
    counter
}

fn get_alternatives(
    raw_question: &FormulaQuestion,
    answer: f32,
    decimals: u8,
    n_alternatives: u8,
) -> (Vec<f32>, u8) {
    let mut alternatives = vec![0 as f32; (n_alternatives - 1) as usize];
    let mut rng = thread_rng();
    let answer_index = rng.gen_range(0, n_alternatives);
    alternatives.insert(answer_index as usize, answer);

    let decimals_pow = 10f32.powi(decimals as i32);
    let delta_min = (raw_question.parameters.get("delta_min").unwrap_or(&1f32)).clone();
    let delta_max = (raw_question.parameters.get("delta_max").unwrap_or(&10f32)).clone();
    let delta_min_decimals = delta_min * decimals_pow;
    let delta_max_decimals = delta_max * decimals_pow + 1f32;
    let range = delta_max_decimals - delta_min_decimals;

    for i in (0..answer_index).rev() {
        let v = (rng.gen_range(0u32, range as u32) as f32) / (decimals_pow as f32);
        alternatives[i as usize] = alternatives[(i + 1) as usize] - (delta_min as f32) - v;
    }
    for i in (answer_index + 1)..n_alternatives {
        let v = (rng.gen_range(0u32, range as u32) as f32) / (decimals_pow as f32);
        alternatives[i as usize] = alternatives[(i - 1) as usize] + (delta_min as f32) + v;
    }

    (alternatives, answer_index)
}

fn get_answer(
    raw_question: &FormulaQuestion,
    variables: &Vec<(String, String)>,
    config: &Config,
) -> f32 {
    let mut eval = String::new();
    for var in variables {
        eval = eval + &format!("{} = {}\n", var.0, var.1);
    }
    let output = "\nprintf('IOLAUS = %.10f', ans)";
    eval = eval + &raw_question.answer + &output;
    let output = Command::new(&config.octave_bin)
        .arg("--no-gui")
        .arg("--no-history")
        .arg("--norc")
        .arg("--no-init-path")
        .arg("--no-window-system")
        .arg("--eval")
        .arg(&eval)
        .output()
        .expect("Não foi possível executar o octave");

    if !output.status.success() {
        panic!("Erro ao executar o octave:\n{}", eval);
    }
    let stdout = String::from_utf8_lossy(&output.stdout);
    let iolaus = match stdout.lines().last() {
        Some(x) => x,
        None => panic!("Erro ao ler última linha: {}", stdout),
    };
    let value_str = iolaus.replace("IOLAUS = ", "");
    let value = match value_str.parse::<f32>() {
        Ok(x) => x,
        Err(_) => panic!("Erro ao interpretar valor: {}", stdout),
    };
    value
}

fn get_variables(raw_question: &FormulaQuestion, decimals: u8) -> Vec<(String, String)> {
    let decimals_pow = 10i32.pow(decimals as u32);
    let mut rng = thread_rng();
    let mut variables = Vec::new();
    for var in &raw_question.variables {
        let name = var.name.clone();
        let min = var.min * decimals_pow;
        let max = var.max * decimals_pow + 1;
        let value = (rng.gen_range(min, max) as f32) / (decimals_pow as f32);
        variables.push((name, format!("{:.*}", decimals as usize, value)));
    }
    variables
}
